// Copyright (c) 2023 The Bitcoin developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

'use strict';
const assert = require('assert');
const config = require('../config');
const {
    parseAliasTx,
    getAliasTxs,
    sortAliasTxsByTxidAndBlockheight,
    getValidAliasRegistrations,
    getAliasStringsFromValidAliasTxs,
} = require('../src/alias');
const { getOutputScriptFromAddress } = require('../src/utils');
const {
    testAddressAliases,
    testAddressAliasesWithUnconfirmedTxs,
    aliases_fake_data,
} = require('./mocks/aliasMocks');

describe('alias-server alias.js', function () {
    it('Correctly parses a 5-character alias transaction', function () {
        const registrationOutputScript = getOutputScriptFromAddress(
            config.aliasConstants.registrationAddress,
        );
        assert.deepEqual(
            parseAliasTx(
                testAddressAliases.txHistory[
                    testAddressAliases.txHistory.findIndex(
                        i =>
                            i.txid ===
                            '9d9fd465f56a7946c48b2e214386b51d7968a3a40d46cc697036e4fc1cc644df',
                    )
                ],
                config.aliasConstants,
                registrationOutputScript,
            ),
            testAddressAliases.allAliasTxs[
                testAddressAliases.allAliasTxs.findIndex(
                    i =>
                        i.txid ===
                        '9d9fd465f56a7946c48b2e214386b51d7968a3a40d46cc697036e4fc1cc644df',
                )
            ],
        );
    });
    it('Correctly parses a 6-character alias transaction', function () {
        const registrationOutputScript = getOutputScriptFromAddress(
            config.aliasConstants.registrationAddress,
        );
        assert.deepEqual(
            parseAliasTx(
                testAddressAliases.txHistory[
                    testAddressAliases.txHistory.findIndex(
                        i =>
                            i.txid ===
                            '36fdab59d25625b6ff3661aa5ab22a4893698fa5618e5e958e1d75bf921e6107',
                    )
                ],
                config.aliasConstants,
                registrationOutputScript,
            ),
            testAddressAliases.allAliasTxs[
                testAddressAliases.allAliasTxs.findIndex(
                    i =>
                        i.txid ===
                        '36fdab59d25625b6ff3661aa5ab22a4893698fa5618e5e958e1d75bf921e6107',
                )
            ],
        );
    });
    it('Returns false for an eToken transaction', function () {
        const registrationOutputScript = getOutputScriptFromAddress(
            config.aliasConstants.registrationAddress,
        );
        assert.deepEqual(
            parseAliasTx(
                testAddressAliases.txHistory[
                    testAddressAliases.txHistory.findIndex(
                        i =>
                            i.txid ===
                            'feafd053d4166601d42949a768b9c3e8ee1f27912fc84b6190aeb022fba7fa39',
                    )
                ],
                config.aliasConstants,
                registrationOutputScript,
            ),
            false,
        );
    });
    it('Returns false for a standard tx without an OP_RETURN', function () {
        const registrationOutputScript = getOutputScriptFromAddress(
            config.aliasConstants.registrationAddress,
        );
        assert.deepEqual(
            parseAliasTx(
                testAddressAliases.txHistory[
                    testAddressAliases.txHistory.findIndex(
                        i =>
                            i.txid ===
                            '7440fb8810610f29197701c53f4a29479a9aede8c66feabb44b049232f990791',
                    )
                ],
                config.aliasConstants,
                registrationOutputScript,
            ),
            false,
        );
    });
    it('Correctly parses all aliases through transactions at test address ecash:qp3c268rd5946l2f5m5es4x25f7ewu4sjvpy52pqa8', function () {
        assert.deepEqual(
            getAliasTxs(testAddressAliases.txHistory, config.aliasConstants),
            testAddressAliases.allAliasTxs,
        );
    });
    it('Correctly parses all aliases through transactions at test address ecash:qp3c268rd5946l2f5m5es4x25f7ewu4sjvpy52pqa8 including unconfirmed txs', function () {
        assert.deepEqual(
            getAliasTxs(
                testAddressAliasesWithUnconfirmedTxs.txHistory,
                config.aliasConstants,
            ),
            testAddressAliasesWithUnconfirmedTxs.allAliasTxs,
        );
    });
    it('Correctly sorts simple template alias txs including unconfirmed alias txs by blockheight and txid', function () {
        assert.deepEqual(
            sortAliasTxsByTxidAndBlockheight(aliases_fake_data.unsortedSimple),
            aliases_fake_data.sortedSimple,
        );
    });
    it('Correctly sorts template alias txs including unconfirmed alias txs by blockheight and txid', function () {
        assert.deepEqual(
            sortAliasTxsByTxidAndBlockheight(aliases_fake_data.allAliasTxs),
            aliases_fake_data.allAliasTxsSortedByTxidAndBlockheight,
        );
    });
    it('Correctly sorts alias txs including unconfirmed alias txs by blockheight and txid', function () {
        assert.deepEqual(
            sortAliasTxsByTxidAndBlockheight(
                testAddressAliasesWithUnconfirmedTxs.allAliasTxs,
            ),
            testAddressAliasesWithUnconfirmedTxs.allAliasTxsSortedByTxidAndBlockheight,
        );
    });
    it('Correctly returns only valid alias registrations at test address ecash:qp3c268rd5946l2f5m5es4x25f7ewu4sjvpy52pqa8', function () {
        assert.deepEqual(
            getValidAliasRegistrations([], testAddressAliases.allAliasTxs),
            testAddressAliases.validAliasTxs,
        );
    });
    it('Correctly returns only new valid alias registrations at test address ecash:qp3c268rd5946l2f5m5es4x25f7ewu4sjvpy52pqa8 given partial txHistory and list of registered aliases', function () {
        // Take only txs after registration of alias 'bytesofman'
        const unprocessedAliasTxs = testAddressAliases.allAliasTxs.slice(
            testAddressAliases.allAliasTxs.findIndex(
                i => i.alias === 'bytesofman',
            ),
        );
        // Get list of all valid alias registrations before 'bytesofman'
        const registeredAliases = getAliasStringsFromValidAliasTxs(
            testAddressAliases.validAliasTxs.slice(
                0,
                testAddressAliases.validAliasTxs.findIndex(
                    i => i.alias === 'bytesofman',
                ),
            ),
        );

        // newlyValidAliases will be all the valid alias txs registered after 'bytesofman'
        const newlyValidAliases = testAddressAliases.validAliasTxs.slice(
            testAddressAliases.validAliasTxs.findIndex(
                i => i.alias === 'bytesofman',
            ),
        );

        assert.deepEqual(
            getValidAliasRegistrations(registeredAliases, unprocessedAliasTxs),
            newlyValidAliases,
        );
    });
    it('Correctly returns valid alias registrations at test address ecash:qp3c268rd5946l2f5m5es4x25f7ewu4sjvpy52pqa8 given some unconfirmed txs in history', function () {
        assert.deepEqual(
            getValidAliasRegistrations(
                [],
                testAddressAliasesWithUnconfirmedTxs.allAliasTxs,
            ),

            testAddressAliasesWithUnconfirmedTxs.validAliasTxs,
        );
    });
    it('getAliasStringsFromValidAliasTxs returns an array of string of the alias object key from an array of valid alias registrations', function () {
        assert.deepEqual(
            getAliasStringsFromValidAliasTxs(
                testAddressAliasesWithUnconfirmedTxs.validAliasTxs,
            ),
            testAddressAliasesWithUnconfirmedTxs.validAliasStrings,
        );
    });
});
